(function() {
	var config = window.offlineAppConfig || {},
		
 	api = {
		randomId: function() {
			return "op_" + Math.random().toString(36).substr(2,9);
		},
		
		makeObjCCall: function(call, query, callback) {
			var url, frame, glob, top;
			url = '--op-objc-' + call + '?';
			for (var k in query) {
				var v = query[k];
				if (typeof v === "object") {
					for (var i in v) {
						url += k + '=' + encodeURIComponent(v[i]) + '&';
					}
				} else {
					url += k + '=' + encodeURIComponent(v) + '&';
				}
			}
			glob = api.randomId();
			top = window.top;
			top[glob] = function(ret) {
				setTimeout(function() {
					frame.parentNode.removeChild(frame);
					delete top[glob];
					if (typeof callback !== 'undefined') {
						callback(ret);
					}
				}, 100);
			};
			url += 'callback=' + glob;
 			setTimeout(function() {
				var d = top.document;
				frame = d.createElement('IFRAME');
				frame.setAttribute('src', url);
				frame.setAttribute('style', 'display: none');
				d.body.appendChild(frame);
			}, 100);
		},
		
		resolveUrl: function(url, callback) {
			api.makeObjCCall('resolve-urls', { url: url }, function(r) {
				callback(typeof url === "string" ? r[0] : r);
			});
		},
		
		revertUrl: function(url, callback) {
			api.makeObjCCall('revert-urls', { url: url }, function(r) {
				callback(typeof url === "string" ? r[0] : r);
			});
		},
		
		parseQueryString: function(s) {
			var pairs = s.split('&'), q = { }, i, p;
			for (i = 0; i < pairs.length; i ++) {
				p = pairs[i].split('=');
				if (p.length >= 2) {
					q[p[0]] = decodeURIComponent(p[1]);
				}
			}
			return q;
		},
		
		listAssets: function(args, callback) {
 			if (typeof args['url'] === "undefined") {
				args['url'] = document.location.href;
 			}
			api.makeObjCCall('list-assets', args, callback);
		},
		
		checkReachability: function(callback) {
			api.makeObjCCall('check-reachability', {}, callback);
		},
		
		passThrough: function(callback) {
 			api.makeObjCCall('pass-through', {}, callback);
		}
	},
	
	clientInit = function() {
		if (typeof window.offlinePagesApi === "object" &&
			typeof window.offlinePagesApi.__bootstrap === "function") {
			window.offlinePagesApi.__bootstrap(api);
			delete window.offlinePagesApi.__bootstrap;
		}
		if (typeof config.init === "function") {
			config.init.call(api);
		}
	},
	
	formsInit = function() {
		jQuery('form').submit(function(e) {
			var form = jQuery(this);
			api.makeObjCCall('form-submit', {
				url: form.attr('action'),
				type: form.attr('method'),
				data: form.serialize()
			});
		});
	},
	
	facadeXhr = function() {
		return {
			readyState: 0,
			
			getResponseHeader: function() { return null; },
			
			getAllResponseHeaders: function() { return null; },
			
			setRequestHeader: function() { return this; },
			
			overrideMimeType: function() { return this; },
			
			abort: function() { return this; },
			
			internalLoopBack: function(o, data) {
				this.readyState = 4;
				this.status = 200;
				this.statusText = "";
				if (typeof o.success === "function") {
					//TODO: bug - support arrays
					o.success(data, "", this);
				}
				if (typeof o.complete === "function") {
					//TODO: bug - support arrays
					o.complete(this, "");
				}
				this.done = function() {};
				this.fail = function(f) {
					f(this, "abort");
				}
				this.then = function(d, f) {
					f(this, "abort");
				}
				this.always = this.fail;
				this.success = this.done;
				this.error = this.fail;
				this.complete = this.always;
				return this;
			},
			
			internalAbort: function(o, handlers) {
				this.readyState = 4;
				this.status = 0;
				this.statusText = "";
				if (typeof o.error === "function") {
					//TODO: bug - support arrays
					o.error(this, "abort");
				}
				if (typeof o.complete === "function") {
					//TODO: bug - support arrays
					o.complete(this, "abort");
				}
				this.done = function() {};
				this.fail = function(f) {
					f(this, "abort");
				}
				this.then = function(d, f) {
					f(this, "abort");
				}
				this.always = this.fail;
				this.success = this.done;
				this.error = this.fail;
				this.complete = this.always;
				
				if (typeof handlers === "object") {
					for (var e in handlers) {
						jQuery.each(handlers[e], function(i, args) {
							switch (e) {
							case "fail":
							case "error":
							case "complete":
							case "always":
								if (typeof args[0] === "function") {
									args[0](this, "abort");
								}
								break;
							case "then":
								if (typeof args[0] === "function" &&
									typeof args[1] === "function") {
									args[1](this, "abort");
								}
								break;
							}
						});
					}
				}
				return this;
			}
		};
	},
	
	ajaxInit = function() {
		var jQuery_ajax = jQuery.ajax;
		jQuery.ajax = function(url, options) {
			var target, callargs, o, xhr, fakeXhr, handlers, data;
			if (typeof url === "string") {
				target = url;
				o = options;
			} else if (typeof url === "object") {
				target = url.url;
				o = url;
			} else {
				target = options.url;
				o = options;
			}
			if (typeof target === "undefined" ||
				target.substr(0,2) === "--") {
				return jQuery_ajax(url, options);
			}
			if (o.data)
			{
				if (typeof o.data !== "string") {
					o.data = jQuery.param(o.data);
				}
				if (typeof o.type !== "string" || o.type.toLowerCase() !== "post") {
					target += "?";
					target += o.data;
				}
			}
			o.url = target;
	
			//TODO: bug - support beforeSend. Careful, e.g. should it be called in context of 
			// original (http:) or modified (file:) request?
			//TODO: properly support statusCode map of callbacks
			
			if (typeof config.ajaxBeforeResolve == "function")
			{
				data = config.ajaxBeforeResolve.call(o);
				if (typeof data === "object") {
					return facadeXhr().internalLoopBack(o, data);
				}
				if (data === false) {
					return facadeXhr().internalAbort(o);
				}
			}
			
			if (typeof o.headers === "object" && o.headers["X-OfflinePages-PassThrough"] === "YES") {
				return jQuery_ajax(o);
			}
			
			//TODO: bug - callbacks are not called in order they were registered??
			
			handlers = {};
			jQuery.each([ "success", "error", "complete" ], function(i, e) {
				if (o[e]) {
					//TODO: bug? o[e] can be an array as of jQuery 1.5
					handlers[e] = [[o[e]]];
					delete o[e];
				}
			});
		
			callargs = {
				url: o.url
			};
			if (typeof o.type === "string" && o.type.toLowerCase() === "post") {
				callargs['type'] = "POST";
				callargs['body'] = btoa(o.data);
			}
			jQuery.extend(callargs, o.offlineData);
			
			api.makeObjCCall('ajax', callargs, function(target) {
				if (typeof target === "undefined") {
					fakeXhr.internalAbort(o, handlers);
					return;
				}
				
				o.url = target;
			
				if (typeof config.ajaxBeforeDownload == "function") {
					config.ajaxBeforeDownload.call(o);
				}
			
				xhr = jQuery_ajax(o).always(function(data, textStatus, xhr) {
					jQuery.each([ "status", "statusCode", "statusText", "responseXML", "responseText", "responseJson", "readyState" ], function(i, p) {
						if (typeof xhr[p] !== "undefined") {
							fakeXhr[p] = xhr[p];
						}
					});
				
					if (typeof config.ajaxBeforeCallbacks == "function") {
						config.ajaxBeforeCallbacks.call(o, fakeXhr, xhr);
					}
				});
				xhr.success = xhr.success || xhr.done;
				xhr.error = xhr.error || xhr.fail;
				xhr.complete = xhr.complete || xhr.always;
				for (var e in handlers) {
					jQuery.each(handlers[e], function(i, f) {
						xhr[e].apply(xhr, f);
					});
				}
				handlers = {};
			});
			
			fakeXhr = facadeXhr();
	
			jQuery.each([ "done", "fail", "always", "success", "error", "complete", "then", "statusCode" ], function(i, e) {
				fakeXhr[e] = function() {
					if (xhr) {
						xhr[e].apply(xhr, arguments);
					} else {
						handlers[e] = handlers[e] || [];
						handlers[e].push(arguments);
					}
					return this;
				};
			});
			return fakeXhr;
		}
	},
	
	ajaxSubmitInit = function() {
		var jQuery_ajaxSubmit = jQuery.fn.ajaxSubmit;
		if (typeof jQuery_ajaxSubmit !== "undefined") {
			jQuery.fn.ajaxSubmit = function(options) {
				if (typeof options == 'function') {
					options = { success: options };
				}
				else if ( options === undefined ) {
					options = { };
				}
				options.offlineData = { isAjaxSubmit: true };
				jQuery_ajaxSubmit.call(this, options);
			};
		}
	},
	
	tryInit = function() {
		try {
			if (typeof jQuery === "undefined") {
				if (document.readyState === "complete") {
					clientInit();
				} else {
					setTimeout(tryInit, 111);
				}
				return;
			}
			
 			if (config.overrideAjaxSubmit === true) {
 				ajaxSubmitInit();
 				config.overrideAjax = true;
 			}
			if (config.overrideAjax !== false) {
				ajaxInit();
			}
			if (config.overrideForms === true) {
				formsInit();
			}
			
			clientInit();
		} catch (e) { }
	},
	
	handleMessage = function(e) {
 		if (typeof e.data === "string") {
			var params = api.parseQueryString(e.data);
			if (params["action"] === "redirect_source") {
				e.source.location = params["url"];
			}
 		}
	};
	
 	if (config.handleMessages !== false) {
		window.addEventListener("message", handleMessage, false);
	}
	
	tryInit();
})();
